from __future__ import print_function
from sys import platform
from os import system
import WalabotAPI as wlbt
import time
import csv
from AWSIoTPythonSDK.MQTTLib import AWSIoTMQTTShadowClient
import sys
import time
import json
import os
import pygame
from pygame.locals import *


# These are my AWS IoT login and certificates
host = "a325207ufsi6fz.iot.us-west-2.amazonaws.com"
cert_path = os.path.realpath(__file__).rstrip(os.path.basename(__file__)) + "cert/"
rootCAPath = cert_path + "rootca.pem"
certificatePath = cert_path + "cert.crt"
privateKeyPath = cert_path + "priv.key"
shadowClient = "DroneRacePracticeCompanion"

# Globals
Power_Status = ["OFF", "ON"]
Power_StatusSelected = Power_Status.index("OFF")
Mode = "None"
lpf_data = 0.0
LapCount = 0
LapTime = 0
BestTime = 0
wlbt_enable = 0
TimeAttackTimer = 0
RaceSeconds = 150 #units is seconds, 2:30 = 150 seconds
RaceEnding = True

# Sound files
snd_err = "Audio/error.wav"
snd_lap = "Audio/lap_complete.wav"
snd_start = "Audio/start_tones.wav"
snd_end = "Audio/end_tones.wav"

# Walabot setup
wlbt_enable = 1
wlbt.Init()                 # load the WalabotSDK to the Python wrapper
wlbt.SetSettingsFolder()    # set the path to the essetial database files
wlbt.ConnectAny()           # establishes communication with the Walabot
wlbt.SetProfile(wlbt.PROF_SENSOR)  # set scan profile out of the possibilities
wlbt.SetDynamicImageFilter(wlbt.FILTER_TYPE_NONE)  # specify filter to use
wlbt.Start()                # starts Walabot in preparation for scanning
print('Walabot Version = {}'.format(wlbt.GetVersion()))

#Walabot Calibration
print("Calibration Start")
wlbt.StartCalibration()
while wlbt.GetStatus()[0] == wlbt.STATUS_CALIBRATING:
    wlbt.Trigger()
print("Calibration Complete")

#init pygame for sound playing and local display
pygame.mixer.init()
pygame.display.init()
pygame.display.set_caption('Drone Race Practice Companion')
size = (pygame.display.Info().current_w, pygame.display.Info().current_h) #rpi touchscreen is 800 x 480
print (size)
screen = pygame.display.set_mode(size, pygame.FULLSCREEN)
screen.fill((0,0,0))    #black
pygame.font.init()      #initialize font

def Walabot_Init():
    global log, basetime, lasttime, datamin, LapCount, lapstart, laptimer, laptrigger
    global laptrigger_n1, lapHighThreshold, lapLowThreshold, f, wlbt_enable
    
    # Walabot Calibration
    print("Calibration Start")
    wlbt.Trigger()
    wlbt.StartCalibration()
    while wlbt.GetStatus()[0] == wlbt.STATUS_CALIBRATING:
        wlbt.Trigger()
    print("Calibration Complete")

    # data logging for debug purposes
    print("Opening Logging File")
    f = open('log.csv','wb')
    log = csv.writer(f, delimiter=',')
    log.writerow(['time','datamin','lpf_data','RawEnergy','CalEnergy','Total Energy','Sample Rate','Lap Count','Lap Timer'])

    # variables for calculating sampling rate
    basetime = time.time()
    lasttime = time.time() - basetime

    # datamin is used to correct for initially high power level reported by Walabot
    datamin = 10

    # lap triggering, counting, and timing variables
    LapCount = 0
    lapstart = time.time()
    laptimer = 0
    laptrigger = 0
    laptrigger_n1 = 0
    lapHighThreshold = 0.06  #may need tweaking for other gates / setups
    lapLowThreshold = 0.03   #may need tweaking for other gates / setups

def IoT_Change_Power(ShadowPayload):
    global Power_StatusSelected, log, f, wlbt_enable, Mode
    
    # Turn on power
    if (ShadowPayload == "ON" and Power_Status[Power_StatusSelected] == "OFF"):
        # Assemble JSON payload
        JSONPayload = '{ "state" : {'+\
                            '"reported": {'+\
                                '"Power": "' + Power_Status[Power_StatusSelected] + '" '+\
                            '}, '+\
                            '"desired": {'+\
                                '"Power": "' + Power_Status[Power_StatusSelected] + '" '+\
                            '} '+\
                        '} '+\
                    '}'
        # Update thing shadow
        myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5) 
        print("Power = " + Power_Status[Power_StatusSelected])

        # Turn on Walabot
        Walabot_Init()
        Power_StatusSelected = Power_Status.index("ON")
        print('power is set to on')

    # Turn off power
    elif (ShadowPayload == "OFF" and Power_Status[Power_StatusSelected] == "ON"):
        Power_StatusSelected = Power_Status.index("OFF")
        # change modes so thing shadow indicates previous mode
        if (wlbt_enable == 1):
            if (Mode == 'TimeAttack'):
                Mode = 'TimeAttackComplete'
            if (Mode == 'FreePractice'):
                Mode = 'FreePracticeComplete'
            # Close the log file                
            f.close()
            wlbt_enable = 0
        # Assemble JSON payload
        JSONPayload = '{ "state" : {'+\
                            '"reported": {'+\
                                '"Mode": "' + Mode + '", '+\
                                '"Power": "' + Power_Status[Power_StatusSelected] + '" '+\
                            '}, '+\
                            '"desired": {'+\
                                '"Mode": "' + Mode + '", '+\
                                '"Power": "' + Power_Status[Power_StatusSelected] + '" '+\
                            '} '+\
                        '} '+\
                    '}'
        # Update thing shadow
        myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5)
        print("Power = " + Power_Status[Power_StatusSelected])
        print('power is set to off')
        
def IoT_Change_LapCount(ShadowPayload):
    global Power_StatusSelected, LapCount, LapTime, BestTime
    
    # Pull in lap count from thing shadow
    try:
        LapCount = int(ShadowPayload)
    except:
        print('failed to set LapCount to {}\n'.format(ShadowPayload))
    # Assemble JSON payload
    JSONPayload = '{ "state" : {'+\
                        '"reported": {'+\
                            '"LapCount": "' + str(LapCount) + '" '+\
                        '} '+\
                    '} '+\
                '}'
    # Update thing shadow
    myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5)
    print("LapCount = " + str(LapCount))

def IoT_Change_LapTime(ShadowPayload):
    global Power_StatusSelected, LapCount, LapTime, BestTime
    
    # Pull in lap time from thing shadow
    try:
        LapTime = int(ShadowPayload)
    except:
        print('failed to set LapTime to {}\n'.format(ShadowPayload))
    # Assemble JSON Payload
    JSONPayload = '{ "state" : {'+\
                        '"reported": {'+\
                            '"LapTime": "' + str(LapTime) + '" '+\
                        '} '+\
                    '} '+\
                '}'
    # Update thing shadow
    myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5)
    print("LapTime = " + str(LapTime))
    
def IoT_Change_BestTime(ShadowPayload):
    global Power_StatusSelected, LapCount, LapTime, BestTime
    
    # Pull in best time from thing shadow
    try:
        BestTime = int(ShadowPayload)
    except:
        print('failed to set BestTime to {}\n'.format(ShadowPayload))
    # Assemble JSON Payload
    JSONPayload = '{ "state" : {'+\
                        '"reported": {'+\
                            '"BestTime": "' + str(BestTime) + '" '+\
                        '} '+\
                    '} '+\
                '}'
    # Update thing shadow
    myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5) 
    print("BestTime = " + str(BestTime))

def IoT_Change_Mode(ShadowPayload):
    global Power_StatusSelected, LapCount, LapTime, BestTime, Mode, TimeAttackTimer, RaceSeconds
    global snd_start, RaceEnding
    
    # pull in Mode from thing shadow
    try:
        Modetemp = ShadowPayload
    except:
        print('failed to set Mode to {}\n'.format(ShadowPayload))
    # Assemble the JSON payload
    JSONPayload = '{ "state" : {'+\
                        '"reported": {'+\
                            '"Mode": "' + Modetemp + '" '+\
                        '} '+\
                    '} '+\
                '}'
    # Update thing shadow
    myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5) #Send the new status as REPORTED values
    print("Mode = " + str(Mode))
    
    # Setup the timer according to the mode
    if (Modetemp == "TimeAttack"):
        Mode = "TimeAttackStart"
    if (Modetemp == "FreePractice"):
        Mode = "FreePracticeStart"

    # Change of mode is the start of a TimeAttack or FreePractice, so get them started.
    if (Mode == "TimeAttackStart" or Mode == "FreePracticeStart"):
        # Begin by calibrating Walabot
        print("Calibration Start")
        wlbt.StartCalibration()
        while wlbt.GetStatus()[0] == wlbt.STATUS_CALIBRATING:
            wlbt.Trigger()
        print("Calibration Complete")
        # Hold off a few seconds, Alexa is still talking
        time.sleep(5)

    # play the start beeps
    if (Mode == "TimeAttackStart" or Mode == "FreePracticeStart"):
        RaceEnding = False
        pygame.mixer.Channel(0).play(pygame.mixer.Sound(snd_start))
        time.sleep(3)

    # Setup the timer according to the mode
    if (Mode == "TimeAttackStart"):
        TimeAttackTimer = time.time() + RaceSeconds
        Mode = "TimeAttack"
    if (Mode == "FreePracticeStart"):
        TimeAttackTimer = time.time()
        Mode = "FreePractice"


# Shadow callback for when a DELTA is received (this happens when Lamda sets a DESIRED value in IoT)
def IoTShadowCallback_Delta(payload, responseStatus, token):
    print(responseStatus)
    payloadDict = json.loads(payload)
    print("++DELTA++ version: " + str(payloadDict["version"]))

    # Desired = POWER change
    if ("Power" in payloadDict["state"]):
        print("Power: " + str(payloadDict["state"]["Power"]))
        IoT_Change_Power(str(payloadDict["state"]["Power"]))

    # Desired = LapCount change
    if ("LapCount" in payloadDict["state"]):
        print("LapCount: " + str(payloadDict["state"]["LapCount"]))
        IoT_Change_LapCount(str(payloadDict["state"]["LapCount"]))
            
    # Desired = LapTime change
    if ("LapTime" in payloadDict["state"]):
        print("LapTime: " + str(payloadDict["state"]["LapTime"]))
        IoT_Change_LapTime(str(payloadDict["state"]["LapTime"]))

    # Desired = BestTime change
    if ("BestTime" in payloadDict["state"]):
        print("BestTime: " + str(payloadDict["state"]["BestTime"]))
        IoT_Change_BestTime(str(payloadDict["state"]["BestTime"]))

    # Desired = Mode change
    if ("Mode" in payloadDict["state"]):
        print("Mode: " + str(payloadDict["state"]["Mode"]))
        IoT_Change_Mode(str(payloadDict["state"]["Mode"]))


# Shadow callback GET for setting initial status
def IoTShadowCallback_Get(payload, responseStatus, token):
    print(responseStatus)
    payloadDict = json.loads(payload)
    print("++GET++ version: " + str(payloadDict["version"]))

    if ("Power" in payloadDict["state"]["desired"]):
        if(str(payloadDict["state"]["reported"]["Power"]).upper() <> str(payloadDict["state"]["desired"]["Power"]).upper()):
            print("Power: " + str(payloadDict["state"]["desired"]["Power"]))
            IoT_Change_Power(str(payloadDict["state"]["desired"]["Power"]))
            
    if ("LapCount" in payloadDict["state"]["desired"]):
        if(str(payloadDict["state"]["reported"]["LapCount"]).upper() <> str(payloadDict["state"]["desired"]["LapCount"]).upper()):
            print("LapCount: " + str(payloadDict["state"]["desired"]["LapCount"]))
            IoT_Change_LapCount(str(payloadDict["state"]["desired"]["LapCount"]))

    if ("LapTime" in payloadDict["state"]["desired"]):
        if(str(payloadDict["state"]["reported"]["LapTime"]).upper() <> str(payloadDict["state"]["desired"]["LapTime"]).upper()):
            print("LapTime: " + str(payloadDict["state"]["desired"]["LapTime"]))
            IoT_Change_LapTime(str(payloadDict["state"]["desired"]["LapTime"]))

    if ("BestTime" in payloadDict["state"]["desired"]):
        if(str(payloadDict["state"]["reported"]["BestTime"]).upper() <> str(payloadDict["state"]["desired"]["BestTime"]).upper()):
            print("BestTime: " + str(payloadDict["state"]["desired"]["BestTime"]))
            IoT_Change_BestTime(str(payloadDict["state"]["desired"]["BestTime"]))


# Shadow callback for updating the AWS IoT
def IoTShadowCallback_Update(payload, responseStatus, token):
    if responseStatus == "timeout":
        print("++UPDATE++ request " + token + " timed out!")
        
    if responseStatus == "accepted":
        payloadDict = json.loads(payload)
        print("++UPDATE++ request with token: " + token + " accepted!")
        if ("desired" in payloadDict["state"]):
            print("Desired: " + str(payloadDict["state"]["desired"]))
        if ("reported" in payloadDict["state"]):
            print("Reported: " + str(payloadDict["state"]["reported"]))

    if responseStatus == "rejected":
        print("++UPDATE++ request " + token + " rejected!")

# Called when Walabot senses that something has flown through the gate
def LapCompleted():
    global Power_StatusSelected, LapCount, LapTime, BestTime, snd_lap, Mode, TimeAttackTimer

    # Only incriment the lap count if TimeAttack or FreePractice modes are active
    if ((Mode == "TimeAttack" ) or (Mode == "FreePractice")):

        # Play the sound to indicate a lap was registered
        pygame.mixer.Channel(1).play(pygame.mixer.Sound(snd_lap))
        LapCount += 1               # incriment lap count
        if (LapTime < BestTime):    # update best lap time
            BestTime = LapTime

        #if time has expired, turn off time attack after final lap completed
        if (Mode == "TimeAttack" and (TimeAttackTimer - time.time()) < 0):
            Mode = "TimeAttackComplete"
    # Assemble JSON payload
    JSONPayload = '{ "state" : {'+\
                        '"reported": {'+\
                            '"Power": "' + Power_Status[Power_StatusSelected] + '", '+\
                            '"LapCount": "' + str(LapCount) + '", '+\
                            '"LapTime": "' + str(LapTime) + '", '+\
                            '"BestTime": "' + str(BestTime) + '", '+\
                            '"Mode": "' + Mode + '" '+\
                        '}, '+\
                        '"desired": {'+\
                            '"Power": "' + Power_Status[Power_StatusSelected] + '", '+\
                            '"LapCount": "' + str(LapCount) + '", '+\
                            '"LapTime": "' + str(LapTime) + '", '+\
                            '"BestTime": "' + str(BestTime) + '", '+\
                            '"Mode": "' + Mode + '" '+\
                        '} '+\
                    '} '+\
                '}'
    # Update thing shadow    
    myDeviceShadow.shadowUpdate(JSONPayload, IoTShadowCallback_Update, 5)
    print('LapCompleted! LapCount={} LapTime={} BestTime={}'.format(LapCount, LapTime, BestTime))

# Draw the local display with the latest information
def DisplayUpdate():
    global TimeAttackTimer, LapCount, BestTime, Mode

    # Assign fonts
    titlefont = pygame.font.SysFont(None,48)
    timefont = pygame.font.SysFont(None,100)
    modefont = pygame.font.SysFont(None,85)

    # Draw strings as viewable items
    #   titles
    Title1 = titlefont.render('Time Remaining', False, (0,0,0))
    Title2 = titlefont.render('Lap Count', False, (0,0,0))
    Title3 = titlefont.render('Best Lap', False, (0,0,0))
    Title4 = titlefont.render('Active Mode', False, (0,0,0))
    #   time remaining
    displaytime = round(TimeAttackTimer - time.time(),2)
    if displaytime < 0:
        displaytime = 0
    Time1 = timefont.render(str(displaytime) + " Seconds", False, (0,0,0))
    #   lap count
    Lap1 = timefont.render(str(LapCount), False, (0,0,0))
    #   mode (is 2 lines)
    displayMode1 = ' '          #mode text line 1
    displayMode2 = ' '          #mode text line 2
    if Mode == 'TimeAttack':
        displayMode1 = 'Time'
        displayMode2 = 'Attack'
    if Mode == 'FreePractice':
        displayMode1 = 'Free'
        displayMode2 = 'Practice'
    Mode1 = modefont.render(displayMode1, False, (0,0,0))
    Mode2 = modefont.render(displayMode2, False, (0,0,0))
    #   best lap time
    displayBestTime = str(round(BestTime,2))
    Best1 = timefont.render(displayBestTime, False, (0,0,0))

    # Pull in the display back ground image    
    background = pygame.image.load('Screen1_DRPC.png')
    screen.blit(background, (0,0))

    # Find positions for all the items that will be drawn
    T1r = Title1.get_rect()
    T1r.centerx = 400
    T1r.centery = 40    
    screen.blit(Title1, T1r)

    T1d = Time1.get_rect()
    T1d.centerx = 400
    T1d.centery = 150    
    screen.blit(Time1, T1d)
    
    T2r = Title2.get_rect()
    T2r.centerx = 136
    T2r.centery = 320    
    screen.blit(Title2, T2r)

    T2d = Lap1.get_rect()
    T2d.centerx = 136
    T2d.centery = 390    
    screen.blit(Lap1, T2d)

    T3r = Title3.get_rect()
    T3r.centerx = 390
    T3r.centery = 320    
    screen.blit(Title3, T3r)

    T3d = Best1.get_rect()
    T3d.centerx = 390
    T3d.centery = 390    
    screen.blit(Best1, T3d)

    T4r = Title4.get_rect()
    T4r.centerx = 650
    T4r.centery = 320    
    screen.blit(Title4, T4r)

    T4d = Mode1.get_rect()
    T4d.centerx = 650
    T4d.centery = 370    
    screen.blit(Mode1, T4d)
    T4d2 = Mode2.get_rect()
    T4d2.centerx = 650
    T4d2.centery = 420    
    screen.blit(Mode2, T4d2)

    # update the display!
    pygame.display.update()

# This is a lowpass alpha filter used to remove noise from the sensed values provided by walabot
def lpf (alpha, newDataPoint, oldDataPoint):
    return alpha * (newDataPoint - oldDataPoint) + oldDataPoint

#main loop, runs continuously 
def loop():
    global log, basetime, lasttime, datamin, LapCount, lapstart, laptimer, laptrigger
    global laptrigger_n1, lapHighThreshold, lapLowThreshold, BestTime, LapTime, TimeAttackTimer
    global RaceEnding, lpf_data, Mode

    if (Power_Status[Power_StatusSelected] == "ON" and (Mode == 'TimeAttack' or Mode == 'FreePractice')):
        # initiates a scan and records signals
        wlbt.Trigger()  

        # variables to keep track of time
        currtime = time.time() - basetime  
        samplerate = 1/(currtime - lasttime)
        lasttime = currtime
        laptimer = time.time() - lapstart
        
        # get the total reflected energy from Walabot
        RawEnergy = wlbt.GetImageEnergy()

        """
        Initially walabot seems to return an artificially high number.
        After it senses something it drops which looks really funny to the detector.
        datamin is the lowest number reported, and subtracting it from RawEnergy creates
        a better looking signal for the detector.  the result is CalEnergy
        """
        if (RawEnergy < datamin):
            datamin = RawEnergy
        CalEnergy = RawEnergy - datamin
        
        """
        Over time the sensed energy floor tend to increase.  Running a lowpass filter on
        CalEnergy returns just that sensed energy floor, which when subtracted from CalEnergy
        gives a nice signal to run through the detector.  the result is TotalEnergy
        """
        lpf_data = lpf(0.1, CalEnergy, lpf_data)
        TotalEnergy = CalEnergy - lpf_data
        
        """
        We determine that a drone has flown in from of walabot if the TotalEnergy number is above
        lapHighThreshold.  To avoid false lap counts it then has to drop below lapLowThreshold before
        another lap can be counted.  laptrigger_n1 is the previous state of laptrigger.
        """
        laptrigger_n1 = laptrigger
        if (TotalEnergy >= lapHighThreshold):
            laptrigger = 1
        if (TotalEnergy <= lapLowThreshold):
            laptrigger = 0
        if (laptrigger == 1 and laptrigger_n1 == 0):
            # A lap has been completed!
            laptimer = time.time() - lapstart
            lapstart = time.time()
            LapTime = laptimer
            # Lap completed will incriment the lapcounter and update the thing shadow with the result
            LapCompleted()
            
        #log data for debug
        if (Power_Status[Power_StatusSelected] == "ON" and (Mode == 'TimeAttack' or Mode == 'FreePractice')):
            log.writerow([currtime,datamin,lpf_data,RawEnergy,CalEnergy,TotalEnergy,samplerate,LapCount,laptimer])

        #If there is only 5 seconds left, play the race end beeps
        if ((TimeAttackTimer - time.time()) <= 5 and RaceEnding == False and Mode == "TimeAttack"):
            RaceEnding = True
            pygame.mixer.Channel(0).play(pygame.mixer.Sound(snd_end))

        #Update the display!
        DisplayUpdate()

"""
Initialization code
"""
DisplayUpdate()

# Connect to AWS IoT thing.  Allows interaction with the thing shadow
myAWSIoTMQTTShadowClient = AWSIoTMQTTShadowClient(shadowClient)
myAWSIoTMQTTShadowClient.configureEndpoint(host, 8883)
myAWSIoTMQTTShadowClient.configureCredentials(rootCAPath, privateKeyPath, certificatePath)
myAWSIoTMQTTShadowClient.configureAutoReconnectBackoffTime(1, 32, 20)
myAWSIoTMQTTShadowClient.configureConnectDisconnectTimeout(10)  # 10 sec
myAWSIoTMQTTShadowClient.configureMQTTOperationTimeout(5)  # 5 sec
myAWSIoTMQTTShadowClient.connect()
myDeviceShadow = myAWSIoTMQTTShadowClient.createShadowHandlerWithName("DroneRacePracticeCompanion", True)

#uncommnet below to create the shadow the first time you are running it
#LapCompleted()    

# Listen on deltas from the IoT Shadow
myDeviceShadow.shadowGet(IoTShadowCallback_Get, 5)
myDeviceShadow.shadowRegisterDeltaCallback(IoTShadowCallback_Delta)

if __name__ == '__main__':
    try:
        print ('DroneRacePracticeCompanion started, Press Ctrl-C to quit.')
        while True:
            #pass
            loop()
            for event in pygame.event.get():
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        pygame.quit()
                        
    finally:
        #myAWSIoTMQTTShadowClient.shadowUnregisterDeltaCallback()
        wlbt.Stop()  # stops Walabot when finished scanning
        wlbt.Disconnect()  # stops communication with Walabot
        #f.close()        
        myAWSIoTMQTTShadowClient.disconnect()
        print ('DroneRacePracticeCompanion stopped.')
        pygame.mixer.Channel(2).play(pygame.mixer.Sound("Audio/error.wav"))
